function zhat = mcnlm_nThreads(y,hr,hs,win,SearchWin,SamplingRatio,nThreads,method)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% zhat = mcnlm_nThreads(y,hr,hs,win,SearchWin,SamplingRatio,nThreads,method)
% denoises an image y using Monte-Carlo Non-local Means
%
% Input: y  - noisy image, size of MxN
%        hr - range parameter
%        hs - spatial parameter
%        win - radius of patch size
%        SearchWin - radius of search neighborhood
%        Sampling ratio - a number in [0,1]
%        nThreads - number of threads for parallel computing
%        method   - one of the following options:
%                   'oracle','uniform','spatial','intensity'
% 
% Output: zhat - denoised image
%
% Stanley Chan
% Harvard University
% Jul 23, 2013
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ymean    = imfilter(y,ones(2*win+1)/(2*win+1)^2,'symmetric');
npca     = 10;
ypca     = pca_projection(y,win,npca);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Switch ON/OFF Parallel Environment
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
mSize = matlabpool('size');  
if mSize==0
    matlabpool(nThreads);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Partition Indices
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
n_total  = numel(y);
n_minor  = mod(n_total,nThreads);
n_major  = n_total - n_minor;

breakpoints = [1:nThreads]*(n_major/nThreads);
jjsets_major = cell(nThreads,1);
for i=1:nThreads
    if i==1  
        jjsets_major{i} = 1:breakpoints(1);
    else
        jjsets_major{i} = breakpoints(i-1)+1:breakpoints(i);
    end
end
jjsets_minor = breakpoints(end)+1:n_total;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main Denoising Step
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
z_major = zeros(n_major/nThreads,nThreads);
switch method
    case 'uniform'
        parfor ThreadID=1:nThreads
            z_major(:,ThreadID) = mcnlm_uniform_sampling(y,ymean,ypca,hr,hs,SearchWin,SamplingRatio,jjsets_major{ThreadID});
        end
        if ~isempty(jjsets_minor)
            z_minor = mcnlm_uniform_sampling(y,ymean,ypca,hr,hs,SearchWin,SamplingRatio,jjsets_minor);
        else
            z_minor = [];
        end
        
    case 'spatial'
        parfor ThreadID=1:nThreads
            z_major(:,ThreadID) = mcnlm_spatial_sampling(y,ymean,ypca,hr,hs,SearchWin,SamplingRatio,jjsets_major{ThreadID});
        end
        if ~isempty(jjsets_minor)
            z_minor = mcnlm_spatial_sampling(y,ymean,ypca,hr,hs,SearchWin,SamplingRatio,jjsets_minor);
        else
            z_minor = [];
        end
        
    case 'oracle'
        parfor ThreadID=1:nThreads
            z_major(:,ThreadID) = mcnlm_oracle_sampling(y,ymean,ypca,hr,hs,SearchWin,SamplingRatio,jjsets_major{ThreadID});
        end
        if ~isempty(jjsets_minor)
            z_minor = mcnlm_oracle_sampling(y,ymean,ypca,hr,hs,SearchWin,SamplingRatio,jjsets_minor);
        else
            z_minor = [];
        end

    case 'intensity'
        parfor ThreadID=1:nThreads
            z_major(:,ThreadID) = mcnlm_intensity_sampling(y,ymean,ypca,hr,hs,SearchWin,SamplingRatio,jjsets_major{ThreadID});
        end
        if ~isempty(jjsets_minor)
            z_minor = mcnlm_intensity_sampling(y,ymean,ypca,hr,hs,SearchWin,SamplingRatio,jjsets_minor);
        else
            z_minor = [];
        end
        
    otherwise
        error('Please input the correct method. \n');
end
        


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Reshape Output
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
zhat = [z_major(:); z_minor(:)];
zhat = reshape(zhat,size(y));