/* Monte Carlo Non-Local Means
 * Version: Uniform Sampling
 *
 * Example:
 * [zhat ActualRatio] = mcnlm_uniform_sampling(y,ymean,ypca,hr,hs,SearchWin,SamplingRatio);
 *
 * Input:
 * y     - noisy image
 * ymean - mean of noisy image; ymean = imfilter(y, ones(2*win+1)/(2*win+1)^2, 'symmetric');
 * ypca  - pca projected image; ypca = pca_projection(y,win,npca);
 * hr    - range parameter; hr = 1.3*noise_sigma/255
 * hs    - spatial parameter; hs = SearchWin/3;
 * SearchWin - Radius of search neighborhood; SearchWin = 10;
 * SamplingRatio - between [0,1]; SamplingRatio = 0.5;
 *
 * Output: 
 * zhat  - denoised image
 * ActualRatio - actual sampling ratio 
 * 
 * Reference: S. Chan, T. Zickler, Y. Lu, "Monte Carlo Non-local Means", submitted to IEEE Trans. Image Process. 2013.
 *
 * Author:
 * Stanley Chan @ Harvard
 * Cretae: Jul 23, 2013
 */

#include <stdio.h>
#include "mex.h"
#include <math.h>
#include "matrix.h"

#undef		MIN
#undef		MAX
#define		MIN(a, b)       ((a) < (b) ? (a) : (b))	
#define		MAX(a, b)       ((a) > (b) ? (a) : (b))

void main_stuff(double *zhat, double *y, double *ymean, double *ypca, int M, int N, int d, double hr, double hs, int SearchWin, double SamplingRatio, double *jjset, double jjsize, double *ActualRatio)
{
    ///////////////////////////////////////
    // Declare Variables
    ///////////////////////////////////////
    
    int i, j, t, jj;                        // dummy running index
    int ir, jr;                             // reference pixel index
    int ic, jc;                             // current   pixel index
    int ic_min, ic_max, jc_min, jc_max;     // limits of pixel indices
    int Mc, Nc;                             // size of current pixel search neighborhood
    int idx_ref, idx_cur;                   // vectorized pixel indices
            
    double sse_s, sse_r;                    // sum squared error: spatial (sse_s) and range (sse_r) 
    double block_ref, block_cur, block_err; // temporary values of computing sse
    double ws, wr;                          // weights: spatial (ws) and range 
    
    double IIDtmp;                          // dummy variable for random numbers
    double p, w;                            // dummy variable for sampling probability
    double num, den;                        // dummy variable for storing numerator and denominator of NLM
    double pmin = 1e-3;
    
    double* IID = NULL;                     // fixed sampling probability [For speed up. Otherwise define IID on the fly.]
    IID = new double[(2*SearchWin+1)*(2*SearchWin+1)];
    for (j=0; j<2*SearchWin+1; j++)
    for (i=0; i<2*SearchWin+1; i++)
    {
        IID[(j*(2*SearchWin+1))+i] = (double)rand()/RAND_MAX;
    }
    double nu;                              // parameter for sampling probability
    double ActualnSample;
    
//     //////////////////////////////////////////
//     // -- Print Actual sampling ratio    -- //
//     double NSampled = 0.0;
//     for (j=0; j<2*SearchWin+1; j++)
//     for (i=0; i<2*SearchWin+1; i++)
//     {
//         if (IID[(j*(2*SearchWin+1))+i]<=SamplingRatio)
//         NSampled = NSampled + 1;
//     }
//     printf("Actual Sampling Ratio = %3e \n", NSampled/((2*SearchWin+1)*(2*SearchWin+1)));
//     mexEvalString("drawnow;");
//     // -- End Print Actual sampling ratio -- //
//     ///////////////////////////////////////////
    
    
    ///////////////////////////////////////
    // Main Routine
    ///////////////////////////////////////

    // Loop over all image pixels (ir,jr)
    for (t=0;t<jjsize;t++)
    {
            jj = jjset[t];
            jj = jj-1;
            ir = jj % M;
            jr = int(jj/M);

//         ///////////////////////////////////////
//         // -- Print Screen to Monitor --     //
//         if(jr%10==0)
//         {
//             printf("Doing %3d / %3d \n", ir, M);
//             mexEvalString("drawnow;");
//         }
//         // -- End Print Screen to Monitor -- //
//         ///////////////////////////////////////
            
            idx_ref = jr*M + ir; // vectorized index for reference pixel

            ///////////////////////////////////////
            // -- Define search neighborhood -- //
            jc_min = MAX(0,jr-SearchWin);
            jc_max = MIN(N-1,jr+SearchWin);
            ic_min = MAX(0,ir-SearchWin);
            ic_max = MIN(M-1,ir+SearchWin);

            Mc = ic_max - ic_min + 1;
            Nc = jc_max - jc_min + 1;
            // -- End Define search neighborhood -- //
            /////////////////////////////////////////
            
            
            
            
            ///////////////////////////////////////
            // -- Compute weights and Denoise -- //
            ///////////////////////////////////////
            num = 0.0;
            den = 0.0;
            ActualnSample = 0.0;
            for (jc=jc_min; jc<=jc_max; jc++)
            for (ic=ic_min; ic<=ic_max; ic++)
            {
                // Use fixed probability pattern to speed up
                // Otherwise, use prob = (double)rand()/RAND_MAX;
                p = SamplingRatio;
                IIDtmp = (double)rand()/RAND_MAX; //IID[((jc-jr)+SearchWin)*(2*SearchWin+1)+((ic-ir)+SearchWin)];
//                 IIDtmp = IID[((jc-jr)+SearchWin)*(2*SearchWin+1)+((ic-ir)+SearchWin)];
                
                if ((IIDtmp<p)|((jc==jr)&(ic==ir)))   // Ensure center pixel is picked
                {
                    // Step 1(a). compute range weight
                    idx_cur = jc*M + ic; // vectorized index for current pixel
                    
                    sse_r = 0.0;
                    for (i=0; i<d; i++)
                    {
                        block_ref = ypca[idx_ref*d+i];
                        block_cur = ypca[idx_cur*d+i];
                        block_err = block_ref-block_cur;
                        sse_r     = sse_r + block_err*block_err;
                    }
                    wr = -sse_r/(d*2*hr*hr);

                    // Step 1(b). comptue spatial weight
                    sse_s = (ic-ir)*(ic-ir) + (jc-jr)*(jc-jr);
                    ws = -sse_s/(2*hs*hs);

                    // Step 1(c). compute overall weight
                    w = exp(wr+ws);
                    
                    // Step 2. Denoise
                    num = num + w*y[jc*M+ic]/p;
                    den = den + w/p;
                    ActualnSample = ActualnSample + 1;
                }
 
            }//-- End FOR(jc=jc_min; jc<=jc_max; jc++)(ic=ic_min; ic<=ic_max; ic++)
            zhat[t] = num/den;
            ActualRatio[t] = ActualnSample/(Mc*Nc);
            ///////////////////////////////////////////
            // -- END Compute weights and Denoise -- //
            ///////////////////////////////////////////
            
            
    }//-- End FOR(t=0; t<jjsize; t++) 
    
}//-- End Void main_stuff








/////////////////////////
// Interface Function
/////////////////////////
void mexFunction(int nlhs, mxArray *plhs[],
        int nrhs, const mxArray *prhs[])
{

    double *zhat;   // zhat  is the denoised output
    double *y;      // y     is noisy input
    double *ymean;  // ymean is mean of noisy input
    double *ypca;   // ypca  is pca projected data
    int SearchWin;  // radius of search neighborhood
    double hr, hs;  // hs = spatial parameter, hr = range parameter
    int M, N, d;    // M = rows of zhat, N = cols of zhat, d = pca dim.
    double SamplingRatio; // Sampling ratio
    double *jjset;  // a vector listing pixels to be denoised
    int jjsize;     // number of elements in jjset

    double *ActualRatio;
    
    const mwSize *ySize;
    mwSize yDims;
    
    const mwSize *ypcaSize;
    mwSize ypcaDims;

    /////////////////////
    // Inputs
    /////////////////////
    // Input 0 (noisy image)
    yDims = mxGetNumberOfDimensions(prhs[0]);
    ySize = mxGetDimensions(prhs[0]);
    y     = mxGetPr(prhs[0]);

    // Input 1 (patch mean of noisy image)
    ymean = mxGetPr(prhs[1]);
    
    // Input 2 (pca basis)
    ypcaDims = mxGetNumberOfDimensions(prhs[2]);
    ypcaSize = mxGetDimensions(prhs[2]);
    ypca     = mxGetPr(prhs[2]);

    // Input 3 (range parameter)
    hr = mxGetScalar(prhs[3]);
    if (hr<=0)
        mexErrMsgTxt("The parameter hr must be positive. Please retry. \n");
        
    // Input 4 (spatial parameter)
    hs = mxGetScalar(prhs[4]);
    if (hs<=0)
        mexErrMsgTxt("The parameter hs must be positive. Please retry. \n");
            
    // Input 5 (search window radius)
    SearchWin = mxGetScalar(prhs[5]);

    // Input 6 (sampling ratio)
    SamplingRatio = mxGetScalar(prhs[6]);
            
    //Input 7 (pixels to be denoised)
    jjset  = mxGetPr(prhs[7]);
    jjsize = MAX(mxGetM(prhs[7]), mxGetN(prhs[7]));
    
    // Calculate M and N
    M = ySize[0];
    N = ySize[1];

    // Calculate Output Dimensions
    int zDims;
    int zSize[2];
    
    zDims    = 2;
    zSize[0] = jjsize;
    zSize[1] = 1;
    d        = ypcaSize[0]; // dimension of pca basis
    
    ////////////////////
    // Outputs
    ////////////////////
    plhs[0]     = mxCreateNumericArray(zDims, zSize, mxDOUBLE_CLASS, mxREAL);
    zhat        = (double *)mxGetData(plhs[0]);  

    plhs[1]     = mxCreateNumericArray(zDims, zSize, mxDOUBLE_CLASS, mxREAL);
    ActualRatio = (double *)mxGetData(plhs[1]); 
    
    main_stuff(zhat,y,ymean,ypca,M,N,d,hr,hs,SearchWin,SamplingRatio,jjset,jjsize,ActualRatio);
}



