clear all
close all
clc

% This routine plots Fig.5(b)
% Purpose: illustrate Theorem 1 by comparing the theoretical bound
% and the experimental histogram
% 
% Stanley Chan @ Harvard
% May 23, 2013
% Aug 16, 2013

addpath(genpath('../code/'));
addpath(genpath('../utilities/'));

% Signal
n = 10000;
x = linspace(0,1,round(n/4));
f = [0.25*(sin(5.5*x)+1.5) 0.5*ones(1,round(n/4)) 0.25*(cos(2.4*x)+1) 0.25*ones(1,round(n/4))];
f = f';

g = f + random('normal',0,5/255,size(f));
g(g>1) = 1;
g(g<0) = 0;

% Denoising Parameters
h         = 15/255;
win       = 2;
kernel    = make_kernel_pyramid(win,1);
SearchWin = n;
gpad      = padarray(g,[win,0],'symmetric');

% Sampling Probability
p = 0.05*ones(n,1);

% Data
w = compute_weight_1D(gpad, 5001, win, kernel, h, SearchWin);
x = g(:);
w = w(:);

% Statistics of Data
mu_A  = sum(w.*x)/n;
mu_B  = sum(w)/n;
mu    = mu_A/mu_B;

% Trials parameters
max_trial = 1e5;    % We set max_trial = 1e5 in our paper
t_set     = 10.^(linspace(log10(0.0001),log10(0.2),100));
% t_set = 0.01;
T         = length(t_set);

% Initializations
lengthT = length(t_set);
P_expr  = zeros(length(t_set),1);
P_theo  = zeros(length(t_set),1);

E1      = zeros(max_trial,T);

% Main Experiment / Generate 1e5 independent trials
parfor trial = 1:max_trial
    if mod(trial,1000)==0
    fprintf('idx = %3g \n', trial);
    end
    
    I = rand(n,1)<p;
    B = (1/n)*sum(w.*I./p);
    A = (1/n)*sum(w.*x.*I./p);
    
    for idx=1:T
        t = t_set(idx);
        E1(trial,idx) = abs(A/B-mu)-t;
    end
end

% Main Experiment / Generate theoretical curve
for idx = 1:length(t_set)
    fprintf('idx = %3g \n', idx);
    t = t_set(idx);
    P_expr(idx) = nnz(E1(:,idx)>0)/max_trial;
    
    a1  = abs(w.*x-mu*w-w*t);
    a2  = abs(w.*x-mu*w+w*t);
    M1  = max(abs(a1)./p);
    M2  = max(abs(a2)./p);
    
    d1 = (1/n)*sum( a1.^2.*(1-p)./p ) + mu_B*M1*t/6;
    d2 = (1/n)*sum( a2.^2.*(1-p)./p ) + mu_B*M2*t/6;
    
    P_theo(idx) = exp(-(n/2)*(t^2)*(mu_B^2/ d1)) + exp(-(n/2)*(t^2)*(mu_B^2/ d2)) + exp(-sum(p));
end

%=== Display result ===%
figure(1);
plot1 = plot(t_set, P_theo, 'r-o', 'LineWidth', 2); hold on; grid on;
plot2 = plot(t_set, P_expr, 'kx',  'LineWidth', 2);
axis([0 0.2 0 2.2]);

L1 = rectangle('Position',[-0.002,-0.05,0.024,2.1],'LineStyle','-.','EdgeColor',[0.3 0.3 0.3],'LineWidth',3); 
L2 = line([0.021 0.1],[0 0.2],'LineStyle','-.','Color',[0.3 0.3 0.3],'LineWidth',3);
L3 = line([0.021 0.1],[2.05 1.85],'LineStyle','-.','Color',[0.3 0.3 0.3],'LineWidth',3);
xlabel('$\varepsilon$','Interpreter','latex','fontsize',12);
ylabel('$\Pr \left[ \left|Z(\mathbf{p}) - z \right| > \varepsilon \right]$','Interpreter','latex','fontsize',12);
% legend([plot1, plot2], 'Large deviations bounds', 'True probabilities');
legend([plot1, plot2], 'Concentration of measure bounds', 'True probabilities');
title('sampling probability $p_j = 0.05$','Interpreter','latex','fontsize',14);

ax1 = gca;
ax2 = axes('Position',get(ax1,'Position'),...
           'XAxisLocation','top',...
           'YAxisLocation','right',...
           'Color','none',...
           'XColor','k','YColor','k');
set(ax2,'xlim',[0 0.1]);
set(ax2, 'Units', 'normalized', 'Position', [0.35 0.25 0.5 0.5]);
idx = 60;
plot(t_set(1:idx), P_theo(1:idx), 'r-o', t_set(1:idx), P_expr(1:idx), 'kx'); grid on;
hkids = get(gca,'child');
set(hkids(1), 'LineWidth', 2);
set(hkids(2), 'LineWidth', 2);

str = {'Zoom-in view'};
annotation('textbox', [.5 .62, .1, .1],...
           'String', str, 'HorizontalAlignment', 'center');
       
export_fig chan05b.eps -transparent;