clear all
close all
clc

% This routine plots Fig.7
% Purpose: compare optimal sampling, approx sampling and uniform sampling
% on cameraman image
%
% Stanley Chan @ Harvard
% May 23, 2013
% Aug 14, 2013
% Nov 14, 2013

addpath(genpath('../code/'));
addpath(genpath('../utilities/'));

f = im2double(imread('cameraman.tif'));
n = numel(f);

g = f + random('normal',0,15/255,size(f));
g(g>1) = 1;
g(g<0) = 0;

% Denoising Parameters
h         = 15/255;
win       = 2;
kernel    = make_kernel_pyramid(win, 2);
SearchWin = n;
gpad      = padarray(g,[win,win],'symmetric');

% Data
wr = compute_weight_2D(gpad, sub2ind([256,256],75,75), win, kernel, h, SearchWin);
H  = fspecial('gaussian',[501,501],50);
ws = H(250+[-73:256-74],250+[-73:256-74]);
ws = ws/max(ws(:));

w = wr(:).*ws(:);
x = g(:);

mu_A  = sum(w.*x)/n;
mu_B  = sum(w)/n;
mu    = mu_A/mu_B;

%%%%%%%%%%% Oracle Sampling %%%%%%%%%%%%%%%%%%%%%%%
a1 = w(:);
a1(a1<1e-8) = 1e-8;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




%%%%%%%%%%% Gaussian Approximated Sampling %%%%%%%%
a2 = ws(:);
a2(a2<1e-8) = 1e-8;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%% Mean Approximation,  7/17/2013 %%%%%%%%
gmean = imfilter(g, ones(2*win+1)/(2*win+1)^2, 'symmetric');
pmean = g(75+[-win:win], 75+[-win:win]);
pmean = mean(pmean(:));
bI = exp(-(gmean-pmean).^2/(2*h^2));
bI = bI(:); 
a3 = bI;
a3(a3<1e-8) = 1e-8;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%% Mean and Spatial Approximation %%%%%%
a4 = bI(:).*ws(:);
a4(a4<1e-8) = 1e-8;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



% Sampling ratio, defined by number of samples K
k_set = round(linspace(1000,256^2,20));
K  = length(k_set);

max_trial = 1000;    % We set max_trial = 1000 in the paper.
E1 = zeros(max_trial, K);
E2 = zeros(max_trial, K);
E3 = zeros(max_trial, K);
E4 = zeros(max_trial, K);
E5 = zeros(max_trial, K);

k1 = zeros(max_trial, K);
k2 = zeros(max_trial, K);
k3 = zeros(max_trial, K);  
k4 = zeros(max_trial, K);  
k5 = zeros(max_trial, K); 

% Main Experiment
for kid = 1:length(k_set)
    k = k_set(kid);

    t1 = max((1/k)*sum(abs(a1(:))),max(abs(a1))) + sqrt(eps);
    myfun1 = @(v)objfun(v,t1,a1,k);
    
    t2 = max((1/k)*sum(abs(a2(:))),max(abs(a2))) + sqrt(eps);
    myfun2 = @(v)objfun(v,t2,a2,k);
    
    t3 = max((1/k)*sum(abs(a3(:))),max(abs(a3))) + sqrt(eps);
    myfun3 = @(v)objfun(v,t3,a3,k);
    
    t4 = max((1/k)*sum(abs(a4(:))),max(abs(a4))) + sqrt(eps);
    myfun4 = @(v)objfun(v,t4,a4,k);
    
    tau1 = fzero(myfun1,[0,1e16]);
    p1 = max(abs(a1)/t1, min(1,abs(a1)*tau1));
    
    tau2 = fzero(myfun2,[0,1e16]);
    p2 = max(abs(a2)/t2, min(1,abs(a2)*tau2));

    tau3 = fzero(myfun3,[0,1e16]);
    p3 = max(abs(a3)/t3, min(1,abs(a3)*tau3));

    tau4 = fzero(myfun4,[0,1e16]);
    p4 = max(abs(a4)/t4, min(1,abs(a4)*tau4));
    
    p5 = k/n;

    fprintf('k = %3g/%3g \n', k,n);

    for trial = 1:max_trial
        
        I1 = rand(n,1)<p1;
        B = (1/n)*sum(w.*I1./p1);
        A = (1/n)*sum(w.*x.*I1./p1);
        k1(trial,kid) = nnz(I1);
        E1(trial,kid) = (abs(A/B-mu)/mu);
        
        I2 = rand(n,1)<p2;
        B = (1/n)*sum(w.*I2./p2);
        A = (1/n)*sum(w.*x.*I2./p2);
        k2(trial,kid) = nnz(I2);
        E2(trial,kid) = (abs(A/B-mu)/mu);

        I3 = rand(n,1)<p3;
        B = (1/n)*sum(w.*I3./p3);
        A = (1/n)*sum(w.*x.*I3./p3);
        k3(trial,kid) = nnz(I3);
        E3(trial,kid) = (abs(A/B-mu)/mu);
        
        I4 = rand(n,1)<p4;
        B = (1/n)*sum(w.*I4);
        A = (1/n)*sum(w.*x.*I4);
        k4(trial,kid) = nnz(I4);
        E4(trial,kid) = (abs(A/B-mu)/mu);

        I5 = rand(n,1)<p5;
        B = (1/n)*sum(w.*I5);
        A = (1/n)*sum(w.*x.*I5);
        k5(trial,kid) = nnz(I5);
        E5(trial,kid) = (abs(A/B-mu)/mu);        
    end
end

% Plot results
fontsize = 12;
set(0,'defaultaxesfontsize',fontsize);
set(0,'defaulttextfontsize',fontsize);

fontname = 'Times New Roman';
set(0,'defaultaxesfontname',fontname);
set(0,'defaulttextfontname',fontname);

fontweight = 'normal';
set(0,'defaultaxesfontweight',fontweight);
set(0,'defaulttextfontweight',fontweight);

figure;
plot(mean(k1,1)/n, mean(E1,1), 'r-o', 'LineWidth', 2); hold on;
plot(mean(k2,1)/n, mean(E2,1), 'k:^', 'LineWidth', 2); hold on;
plot(mean(k3,1)/n, mean(E3,1), 'b:+', 'LineWidth', 2, 'MarkerSize', 10); 
plot(mean(k4,1)/n, mean(E4,1), 'k:x', 'LineWidth', 2, 'MarkerSize', 10); 
plot(mean(k5,1)/n, mean(E5,1), 'm-*', 'LineWidth', 2, 'MarkerSize', 10); 
hold off;
ylabel('$E[|Z-z|/|z|]$','interpreter','latex','fontsize',16);
xlabel('$\xi$','interpreter','latex','fontsize',16);
legend('Oracle Sampling','Spatial Approximation','Intensity Approximation','Spatial+Intensity Approximation','Uniform Sampling');
grid on;

export_fig chan07.eps -transparent -RGB;