%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This routine generates data for Table I and II
%
% Variations over 24 sampling realizations
% and 20 noise realizations
%
% Stanley Chan @ Harvard
% May 23, 2013
% Apr 11, 2014
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear all
close all
clc

addpath(genpath('../offshelf_packages/'));
addpath(genpath('../code/'));
addpath(genpath('../utilities/'));

currentFolder = pwd;

% Outer Parameters
SearchWin = 17;
win       = 3;

% Parameters
hs = SearchWin/3;
ws = fspecial('gaussian',[2*SearchWin+1,2*SearchWin+1],hs);
b  = ws(:)/max(ws(:));

% Load data
folder = '../data/standard_images/';
list_fname = dir(folder);

% Sampling Ratio
xi_set = [0.05 0.1 0.2 0.5];

% Noise Variances
sigma_set = [10 20 30 40 50];

xlength = length(xi_set);
slength = length(sigma_set);
flength = 10;

max_trial_s = 24;
max_trial_n = 10;

% Initialize
psnr_full  = zeros(max_trial_n,slength,flength);
psnr_mcnlm = zeros(max_trial_s,max_trial_n,xlength,slength,flength);
psnr_kd    = zeros(max_trial_n,slength,flength);
psnr_am    = zeros(max_trial_n,slength,flength);
psnr_bm3d  = zeros(max_trial_n,slength,flength);

for sidx = 1:5
    sigma = sigma_set(sidx);

    for fid=1:10
        for nidx=1:max_trial_n
            
            fprintf('fid = %3g \n', fid);

            fname = sprintf('%s%s', folder, list_fname(fid+2).name);
            f = im2double(imread(fname));
            if size(f,3)==3
                f = rgb2gray(f);
            end

            rand('seed', f(round(end/2),round(end/2),1) );
            
            g = f + random('normal',0, sigma/255, size(f));
            g(g>1) = 1;
            g(g<0) = 0;

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Full NLM
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            hr      = 1.3*sigma/255;
            fhat1   = mcnlm(g,hr,hs,win,SearchWin,1,'uniform');
            psnr_full(nidx,sidx,fid) = -10*log10(norm(fhat1(:)-f(:))^2/numel(f));

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % KD Tree
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            cd ../offshelf_packages/;
            imwrite(g,'noisy.png');
            hr = 1.3*sigma/255;
            npca = 10;
            patch_diameter = 2*win+1;
            s = (patch_diameter - 1)/6;
            full_cmd = sprintf('ImageStack.exe -load noisy.png -nlmeans %3.4g %3.4g %3.4g %3.4g exact -save output.png',s,npca,hs,hr*sqrt(npca));        
            system(full_cmd);
            fhat3 = im2double(imread('output.png'));
            psnr_kd(nidx,sidx,fid) = -10*log10(norm(fhat3(:)-f(:))^2/numel(f));
            cd(currentFolder);


            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Adaptive Manifold
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            hr = 2*sigma/255;  
            [~,fhat4]          = adaptive_manifold_filter(g, hs, hr); 
            fhat4              = fhat4(:,:,1);
            psnr_am(nidx,sidx,fid)  = -10*log10(norm(fhat4(:)-f(:))^2/numel(f));


            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % BM3D
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            [~, fhat5] = BM3D(1, g, sigma);
            psnr_bm3d(nidx,sidx,fid) = -10*log10(norm(fhat5(:)-f(:))^2/numel(f));


            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Monte Carlo
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            hr = 1.3*sigma/255;
            for xi_idx = 1:length(xi_set)
                xi = xi_set(xi_idx);
                for strial = 1:max_trial_s
                    fhat2 = mcnlm(g,hr,hs,win,SearchWin,xi,'spatial');
                    psnr_mcnlm(strial,nidx,xi_idx,sidx,fid) = -10*log10(norm(fhat2(:)-f(:))^2/numel(f));
                end
            end
        end
    end
    sname = sprintf('../data/Table1_data/change_bothA_win%02d_sigma%02d.mat',SearchWin,sigma);
    save(sname,'psnr_mcnlm','psnr_full','psnr_kd','psnr_am','psnr_bm3d');
end